import os, sys, traceback

# Import of sorts
old_init = __init__
__init__ = False
execfile('slides.py')
__init__ = old_init

import sdl_doc
from const import P
import random
import time
import math
from collections import Counter, namedtuple
from uielem import uidict
execfile('data.py')
execfile('die_shape.py')

def test(icon='spoon_icon'):
    root = doc[doc['selection.root']]
    root.append(Node("point", value=P(100, 100), icon=icon))

def add_icons():
    # No good because circles can't be filled!
    # Better use svg paths then...
    doc['references'].extend(icons())

def import_path(path, **kwargs):
    root = doc[doc['selection.root']]
    root.append(Node("path", children=[
        Node("line", fill_color=fill, p_start=P(*prev), p_end=P(*next_))
        for prev, next_ in zip(path[:-1], path[1:])], **kwargs))

def parse_fill(fill):
    if fill is None:
        return None
    return tuple(int(fill[i+1:i+3], 16) / 255.0 for i in (0, 2, 4))

icon_scale = 0.45

# Problem: Not enough contrasts or lines too thick when small.
def import_group(group, scale=icon_scale, **kwargs):
    root = doc['references']
    root.append(Node("group", children=[
        Node("path", fill_color=parse_fill(fill), #stroke_color=None if fill is not None else (0, 0, 0),
             children=[
            Node("line", p_start=scale*P(*prev), p_end=scale*P(*next_))
            for prev, next_ in zip(path[:-1], path[1:])])
        for fill, path in group], **kwargs))

def import_icon(path, id):
    doc['references'].append(Node("path", id=id, children=[
        Node("line", p_start=P(*prev), p_end=P(*next_))
        for prev, next_ in zip(path[:-1], path[1:])]))

def load_imports():
    #import_path(icon3, id='spoon_icon', fill_color=palette[16], transform=('scale', P(0.25, 0.25)))
    #import_path(icon3, id='largespoon_icon', fill_color=palette[16], transform=('scale', P(0.5, 0.5)))
    # import_group(icons['Bake'], id='bake_icon', transform=('scale', P(0.5, 0.5)))
    for key, value in icons.items():
        import_group(value, id='%s_icon' % key.split()[0].lower())
    for key, value in icons.items():
        import_group(value, scale=icon_scale * 2, id='large%s_icon' % key.split()[0].lower())

def add_face3(icons, transforms=None):
    root = doc[doc['selection.root']]
    root.append(Node("group", transforms=transforms, children=[
        rectangle2(px_topleft=P(-50, -50), px_botright=P(50, 50)),
        Node("point", value=P(0, 0), icon=icon),
        Node("point", value=P(-30, 30), icon=icon),
        Node("point", value=P(30, -30), icon=icon)]))

icon_shape = P(80, 80) * icon_scale
die_shape = icon_shape * 1.6 * 2
    
def add_face(icons, transforms=None, root=None, replace=False):
    if root is None:
        root = doc[doc['selection.root']]
    children=[rectangle2(px_topleft=-1.6 * icon_shape, px_botright=1.6 * icon_shape)] + face_children(icons)
    if replace:
        root.L.clear()
        root.L.extend(children)
        root.L['face_value']=icons
    else:
        root.append(Node("group", face_value=icons, transform=transforms, children=children))

def face_children(icons):
    return [Node("point", value=icon_shape * (shape - P(1, 1)), icon="%s_icon" % icon)
            for icon, shape in zip(icons, shapes[len(icons)])]

def face_grid(icon, width, height):
    for i in range(width):
        for j in range(height):
            add_face(icon, transforms=("translate", P(i, j) * 100))

def set_faces(faces, root=None):
    root.clear()
    for i, face in enumerate(faces):
        add_face(face,
                 transforms=("translate", P(i * 1.1 + 0.5, 0) * die_shape),
                 root=root)

faces = [('spoon', 'spoon', 'skull'), ('spoon',), ('knife',)] + [('bake',)] * 3

def color_palette():
    size = 30
    group = Node("group", transform=("translate", doc['editor.mouse_txy']),
                 button_group=True,
                 skip_points=True,
                 children=[
        rectangle2(px_topleft=P(i*size, 0),
                   px_botright=P(i*size, 0) + P(size, size),
                   fill_color=col,
                   on_click="doc['selection.0.ref.fill_color'] = %s" % (col,),
                   on_shift_click="doc['selection.0.ref.stroke_color'] = %s" % (col,))
                     for i, col in enumerate(palette)])
    doc[doc['selection.root']].append(group)

class Char(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        return 'Char(%s)' % ', '.join('%s=%s' % kv
                                      for kv in self.__dict__.items())


def add_button(name, text, on_click, start=P(0, 0), root=None):
    if root is None:
        root = doc['buttons']
    root.append(Node("group", id='%s_button' % name,
                     transform=("translate", start),
                     on_click=on_click,
                     children=[
                         rectangle2(px_topleft=P(0, 0), px_botright=P(80, 30),
                                    fill_color=(0.9,) * 3),
                         Node("text", value=text, p_botleft=P(10, 30))]))
    end = root.L[-1][1].bbox()[1][0] + 10
    root.L['-1.0.botright.value'] = P(end, 34)
    return end

debug = False

def game_init():
    dice = [[("spoon",)] * 2 + [tuple()] * 2 + [("bake",)] + [('reroll',)] + [("skull", "spoon", "spoon")]]
    doc['player.dice'] = dice
    doc['game.level'] = 1
    player = Char(name="player", food=6, max_food=6, dice=dice)
    for key, value in player.__dict__.items():
        doc['player'][key] = value
    doc['player.prizes'] = []
    doc['game.grid_shape'] = (4, 4)
    if debug:
        doc['player.dice'] = [[('spoon',), ('spoon',), (), (), ('bake',), ('reroll',), ('skull', 'spoon', 'spoon')], [('spoon', 'skull', 'spoon'), ('bake', 'skull', 'knife'), ('spoon',), ('bake',), ('reroll', 'skull', 'spoon'), ('knife', 'skull')]]
        doc['game.level'] = 2

def _roll(dice):
    return [random.choice(die) for die in dice]

def roll():
    doc['game.reroll'] = 3
    doc['adversary.faces'] = _roll(doc['adversary.dice'])
    set_faces(doc['adversary.faces'], doc['adversary_faces'])
    doc['player.faces'] = [None for _ in doc['player.dice']]
    reset_keep_buttons()
    reroll()

def reroll(indices=None):
    if not doc['game.reroll']:
        return
    doc['game.reroll'] -= 1
    if indices is None:
        indices = [i for i, child in enumerate(doc['keep_buttons'])
                   if child['0.fill_color'] == palette[3]]
    for ind in indices:
        doc['player.faces'][ind] = random.choice(doc['player.dice'][ind])
    print 'faces', doc['player.faces']
    print 'adv', doc['adversary.faces']
    set_faces(doc['player.faces'], doc['player_faces'])

skull_faces = ['spoon', 'bake', 'knife', 'reroll', 'spoon', 'blank']

def to_tuple(x):
    return x if isinstance(x, tuple) else\
           (x,) if x != "blank" else tuple()

adversaries1 = {"nom": Char(name="nom", max_food=2,
                            dice=[["blank"] * 2 + ["spoon"] * 3 + [("spoon", "spoon")]],
                            prize=["spoon"]),
                "tank": Char(name="tank", max_food=3,
                            dice=[["spoon"] * 2 + ["bake"] + ["blank"] * 3],
                            prize=['bake']),
               "reroll": Char(name="reroll", max_food=1,
                            dice=[["reroll", "spoon", "blank"] + [("spoon", "spoon")] * 3],
                            prize=['reroll']),
               "knife": Char(name="knife", max_food=2,
                             dice=[["knife"] * 3 + ["spoon", "blank", "blank"]] * 2,
                             prize=['knife'])}

adversaries2 = {"nom": Char(name="nom", max_food=2,
                            dice=[["blank"] + ["spoon"] * 4 + [("spoon", "spoon")]],
                            prize=["spoon"]),
               "tank": Char(name="tank", max_food=3,
                            dice=[["spoon"] * 3 + ["bake"] * 2 + ["blank"]],
                            prize=['bake']),
               "reroll": Char(name="reroll", max_food=1,
                            dice=[["reroll", "spoon", "blank"] + [("spoon", "spoon")] * 3],
                            prize=['reroll']),
               "knife": Char(name="knife", max_food=2,
                             dice=[["knife"] * 3 + ["spoon", "spoon", "blank"]] * 2,
                             prize=['knife'])}

adversaries = adversaries1

for char in adversaries1.values():
    char.dice = [[to_tuple(face) for face in die] for die in char.dice]

def toggle_keep(i):
    fill_color = doc['keep%s_button.0.fill_color' % i]
    if fill_color == palette[12]:
        doc['keep%s_button.0.fill_color' % i] = palette[3]
    else:
        doc['keep%s_button.0.fill_color' % i] = palette[12]

def reset_keep_buttons():
    for button in doc['keep_buttons']:
        button['0.fill_color'] = palette[3]

def add_keep_buttons():
    doc['keep_buttons'].clear()
    for i in range(len(doc['player.dice'])):
        add_button("keep%s" % i, "Keep", "toggle_keep(%s)" % i,
                   start=P(i * 1.1 + 0.5 - 0.15, 0) * die_shape,
                   root=doc['keep_buttons'])
    reset_keep_buttons()

def next_adv(name):
    doc['editor.slide'] = 'encounter'
    add_keep_buttons()
    doc['game.turn'] = 0
    for key, value in adversaries[name].__dict__.items():
        doc['adversary'][key] = value
    doc['adversary.dice'] *= doc['game.level']
    doc['adversary.max_food'] *= doc['game.level']
    doc['adversary.food'] = doc['adversary.max_food']
    doc["keep_buttons"]
    print doc['adversary']
    roll()

def flatten(faces):
    flat_faces = []
    for face in faces:
        if isinstance(face, str):
            flat_faces.append(face)
        else:
            flat_faces.extend(face)
    return flat_faces

def execute(self, opp):
    # TODO: Need to execute skulls and rerolls!!
    faces = Counter(flatten(self.L['faces']))
    print "executing", self.L['name'], faces
    if faces['skull'] == 2:
        self.L['food'] -= 1
    elif faces['skull'] == 3:
        return
    elif faces['skull'] >= 4:
        self.L['food'] -= 66
    opp.L['food'] -= faces['spoon'] + (faces['knife'] / 2) * 3
    self.L['food'] += faces['bake']

def get_prize():
    assert(len(doc['adversary.prize']) == 1)
    prize = (doc['adversary.prize'][0],) * doc['game.level']
    if doc['game.turn'] % 2 == 0:
        prize += ('skull', random.choice(skull_faces))
        if prize[-1] == 'blank':
            prize = prize[:-1]
    print "Got prize", prize
    doc['player.prizes'].append(prize)
    doc['editor.slide'] = 'map'
    add_face(prize, root=doc[doc['adversary.ref']], replace=True)
    del doc[doc['adversary.ref']]['on_click']
    if len(doc['player.prizes']) == 6:
        next_level()

def execute_all():
    doc['game.turn'] += 1
    reroll_adv = Counter(flatten(doc['player.faces']))['reroll'] > 0
    reroll_player = Counter(flatten(doc['adversary.faces']))['reroll'] > 0
    print "Rerolls player=%s adv=%s" % (reroll_player, reroll_adv)
    if reroll_adv:
        doc['adversary.faces'] = _roll(doc['adversary.dice'])
        set_faces(doc['adversary.faces'], doc['adversary_faces'])
    if reroll_player:
        doc['player.faces'] = _roll(doc['player.dice'])
        set_faces(doc['adversary.faces'], doc['adversary_faces'])
    execute(doc['player'], doc['adversary'])
    if doc['adversary.food'] > 0:
        execute(doc['adversary'], doc['player'])
    for char in [doc['player'], doc['adversary']]:
        char.L['food'] = min(char.L['food'], char.L['max_food'])
        if char.L['food'] <= 0:
            print(char.L['name'], "is out of food")
    print("player", doc['player.food'])
    print("adv", doc['adversary.food'])
    if doc['adversary.food'] <= 0:
        get_prize()
    else:
        roll()

def next_level():
    doc['game.level'] += 1
    doc['player.dice'].append(doc['player.prizes'])
    doc['player.prizes'] = []
    doc['player.food'] = doc['player.max_food']
    doc['editor.slide'] = 'map'
    gen_grid(*doc['game.grid_shape'])

def grid_click(id, name):
    next_adv(name)
    doc['adversary.ref'] = id

def gen_grid(width, height, root=None):
    if root is None:
        root = doc['grid']
    root.clear()
    for i in range(width):
        for j in range(height):
            name = random.choice(adversaries.keys())
            icon = {"nom": "spoon", "tank": "bake"}.get(name, name)
            add_face(['large' + icon], root=root,
                     transforms=("translate", P(i + 0.5, j + 0.5) * 1.1 * die_shape))
            root.L[-1]['pos'] = (i, j)
            root.L[-1]['on_click'] = "grid_click('%s', '%s')" % (root.L[-1]['id'], name)
            #transforms={"scale": ("scale", P(0.5, 0.5) * 2),
            #"zmove": ("translate", P(i, j) * 100)})

def run_button():
    root = doc[doc["selection.root"]]
    txy = doc["editor.mouse_txy"]
    for child, transform in root.dfs():
        # Problem: dfs includes child's transform and so does bbox.
        if "on_click" in child and\
           collide(child, txy, transform=transform):
            print "clicked on", child["id"], "on_click" in child
            run_text(child["id"], "on_click")
            return True
    return False

def quit():
    uidict["root"].quit()

if __init__:
    uidict["root"].withdraw()
    doc['drawing'].pop()
    doc['editor.slide'] = 'map'
    doc['drawing'].extend(n("""
    group:
      group: id="custom_ui" visible=True
      group: id="title" visible=exr('`editor.slide == `self.id')
      group: id="map" visible=exr('`editor.slide == `self.id')
        text: value=exr('"level %s" % `game.level') p_botleft=P(20, 70)
        text: value=exr('"food: %s" % `player.food') p_botleft=P(20, 100)
        group: id="grid" visible=True transform=("translate", P(180, 40))
      group: id="encounter" visible=exr('`editor.slide == `self.id')
        group: id="player_faces" transform=("translate", P(20, 350))
        group: id="adversary_faces" transform=("translate", P(20, 120))
        group: id="keep_buttons" transform=("translate", P(0, 420))
        group: id="buttons" transform=("translate", P(20, 495))
        group: id="status"
          text: value=exr("'food: %s' % `player.food") p_botleft=P(20, 485)
          text: value=exr("'food: %s' % `adversary.food") p_botleft=P(20, 40)
    """))
    doc['editor.slides'] = ['title', 'map', 'encounter']
    doc['ui'].extend(n("""
    group:
      group: id="game" reroll=3
      group: id="player"
      group: id="adversary"
    """))
    doc.saved = [doc.m]
    doc.undo_index = -1
    doc['overlay.transforms'] = exr('`drawing.transforms')
    doc['editor.filename'] = 'pyweek27_doc.py'
    #doc.sync()
    #doc.load()
    doc['editor.slide'] = 'map'
    load_imports()
    game_init()
    #doc['buttons'].append(Node("group", id='reroll_button', children=[
    #    rectangle2(px_topleft=P(0, 0), px_botright=P(100, 30), fill_color=(0.1, 0.1, 0.1))]))
    end = add_button("reroll", "Reroll", "reroll()")
    doc['buttons'].append(Node("text", value=exr("`game.reroll"),
                               p_botleft=P(end + 10, 30)))
    end = doc['buttons.-1'].bbox()[1][0] + 2
    add_button("go", "Go", "execute_all()", start=P(end + 10, 0))
    doc['drawing.skip_points'] = True
    gen_grid(*doc['game.grid_shape'])
    # gen_grid(1, 1)

input_callbacks += """
quit = key_press(q) !quit
command = @quit | @button | @zoom | @scroll
"""

print input_callbacks
